#!/usr/bin/env python
import logging
import os
import pathlib
import random
import re
import string
import lxml.html
from urllib3.util import parse_url

from docutils.core import publish_string

from odoo.addons.base.models.ir_module import MyWriter
from odoo.addons.saas_worker.util import zip_dir     # noqa
from odoo.modules.module import MANIFEST_NAMES
from odoo.tools import exception_to_unicode, plaintext2html, cloc
from odoo.osv.expression import get_unaccent_wrapper
from odoo.tools.mail import html_sanitize

from werkzeug.urls import url_parse
from collections.abc import Iterable
from ast import literal_eval
from lxml import etree

_logger = logging.getLogger(__name__)
ALIASES = ['Odoo S.A.', 'Odoo SA', 'Odoo', 'OpenERP S.A.', 'OpenERP SA', 'OpenERP SA.', 'OpenERP s.a.', 'OpenERP sa', 'OpenERP', 'Tiny']
RECENT_THRESHOLD = 30  # Default threshold that defines recent data in days

# Maximum number of download hits tolerated by IP address, per module.
# Once above this threshold, we consider that the user behind the IP tries
# to artificially inflates download counts.
DL_HIT_THRESHOLD = 10

CATEGORIES_APPS = [
    "Accounting",
    "Discuss",
    "Document Management",
    "eCommerce",
    "Human Resources",
    "Industries",
    "Localization",
    "Manufacturing",
    "Marketing",
    "Point of Sale",
    "Productivity",
    "Project",
    "Purchases",
    "Sales",
    "Warehouse",
    "Website",
    "Extra Tools",
    "Tutorial",
]

CATEGORIES_THEMES = [
    "Backend",
    "Corporate",
    "Creative",
    "Ecommerce",
    "Education",
    "Entertainment",
    # "Nonprofit",
    "Personal",
    "Retail",
    "Services",
    "Technology",
    "Miscellaneous",
]

CATEGORIES_INDUSTRY = [
    'Services',
    'Retail',
    'Manufacturing',
    'eCommerce',
    'Public',
    'NGO',
]

DOWNLOADS = 'download_count_all_recent desc, repo_prio desc, shortdesc'
DEFAULT = 'sales_recent desc, ' + DOWNLOADS
ORDERS = {
    'Relevance': 'score desc, ' + DEFAULT,
    'Best Sellers': DEFAULT,
    'Name': 'shortdesc, ' + DEFAULT,
    'Ratings': 'average_rating desc, rating_count desc, ' + DEFAULT,
    'Lowest Price': 'price, ' + DEFAULT,
    'Highest Price': 'price desc, ' + DEFAULT,
    'Downloads': DOWNLOADS,
    'Purchases': 'purchases_count_recent desc, ' + DEFAULT,
    'Newest': 'approved, m.create_date desc, ' + DEFAULT,
}

FORBIDDEN_NETWORK_NAMES = [
    'GOOGLE-CLOUD',
    'AT-88-Z',  # Amazon
    'AMAZO',
]

# Map containing the compatibility between licenses,
# it is constructed as {key} can depend on licenses in {value}
LICENSES_MAP = {
    'AGPL-3': ('AGPL-3', 'GPL-3', 'LGPL-3'),
    'GPL-3': ('GPL-3', 'LGPL-3'),
    'LGPL-3': ('LGPL-3', 'OPL-1', 'OEEL-1', 'Other OSI approved licence', 'Other proprietary'),
    'OPL-1': ('LGPL-3', 'OPL-1', 'OEEL-1', 'Other OSI approved licence', 'Other proprietary'),
    'Other OSI approved licence': ('LGPL-3', 'OEEL-1', 'OPL-1', 'Other OSI approved licence', 'Other proprietary'),
    'Other proprietary': ('LGPL-3', 'OEEL-1', 'OPL-1', 'Other OSI approved licence', 'Other proprietary'),
    'OEEL-1': ('LGPL-3', 'OEEL-1'),
}


class DependencyError(ValueError):
    pass

class ManifestError(ValueError):
    pass

class CheckURLError(ValueError):
    pass

class ModuleRepoError(NameError):
    pass

class VCSError(RuntimeError):
    pass


def clean_path(path):
    return os.path.realpath(os.path.expanduser(path))

def get_changelog(path):
    try:
        changelog_file = os.path.join(path, 'doc', 'changelog.rst')
        if os.path.isfile(changelog_file):
            with open(changelog_file) as cl:
                return cl.read()
    except IOError:
        # file not found, don't care
        return None

def rst2html(content, catch=True):
    overrides = dict(embed_stylesheet=False, doctitle_xform=False, output_encoding='unicode',
                     xml_declaration=False, file_insertion_enabled=False, raw_enabled=False)
    try:
        rst = publish_string(source=content, settings_overrides=overrides, writer=MyWriter())
        return html_sanitize(rst, silent=False)
    except Exception as e:
        if not catch:
            raise
        _logger.exception('failed to render rst')
        return "Hmmm, that's embarrassing. Failed to render rst: %s" % (exception_to_unicode(e),)

def html_sanitize_hyperlinks(content, allowed_hrefs=[]):
    assert type(content) == str
    if not content:
        return content
    html = lxml.html.document_fromstring(content)
    if html is not None and len(html):
        for element, attribute, link, pos in html.iterlinks():
            base_href = element.get('href')
            if base_href and base_href != '#' and not any(href in base_href for href in allowed_hrefs):
                element.tag = 'span'
                element.set('href', '')
        return lxml.html.tostring(html)
    return content

def chatter_log(record, level, notify=True):
    init = record.env.registry._init
    def wrap(msg, *args, **kwargs):
        if init:
            return
        getattr(_logger, level)(msg, *args, **kwargs)
        subtype = ('loempia.mt_%s' % (level,)) if level != 'info' else None
        # all messages are sent in plaintext -> we convert it manually (12.0)
        msg = plaintext2html(msg)
        if notify:
            record.message_post(body=msg % args, message_type='notification', subtype_xmlid=subtype)
        else:
            record._message_log(body=msg % args)
    return wrap

def search_for_modules(env, module_type, category="", maintainer_id=None, limit=None, **post):
    order = post.get('order', 'Best Sellers').replace('+', ' ')
    if order == 'Latest Updated':
        order = 'Newest'
    order = ORDERS[order]
    search = [
        "NOT (m.is_hidden OR m.is_deleted) AND (m.status = 'approved' OR m.status ='invalid')\
         AND NOT m.score = 0 AND NOT m.shortdesc ilike '%%debrand%%'\
         AND NOT m.name ilike '%%debrand%%' AND NOT m.summary ilike '%%debrand%%'"
    ]
    join_clauses = []
    args = {}
    #todo refactor properly for the unaccent feature
    unaccent = get_unaccent_wrapper(env.cr)
    search += ['m.is_theme'] if 'themes' in module_type else ['NOT m.is_theme']

    if isinstance(category, int):
        search += ["m.category_id = %(category_id)s"]
        args['category_id'] = category
    elif category not in ["", "All"] or module_type == 'industries':
        join_clauses += ["JOIN ir_module_category c ON c.id = m.category_id"]
        if module_type == 'industries' and not category:
            search += ["c.name->>'en_US' similar to %(industries)s"]
            args['industries'] = "({})".format('|'.join(CATEGORIES_INDUSTRY))
        if category not in ["", "All"]:
            search += ["c.name->>'en_US' ilike %(category)s"]
            args['category'] = "%{}%".format(category)

    if post.get('price') in ['Free', 'Paid']:
        search += ["m.price <= 0.0"] if post.get("price") == 'Free' else ["m.price > 0.0"]

    if post.get('data_module'):
        search += ["m.is_data_module"]
        search += ["m.approved"]

    if post.get("approval") in ['Certified', 'Community']:
        search += ["m.approved"] if post.get("approval") == 'Certified' else ["not m.approved"]

    if post.get("author"):
        author, author_srch = map(unaccent, ['m.author', '%(author)s'])
        search += ["{author} ilike {author_srch}".format(
            author=author,
            author_srch=author_srch
        )]
        args['author'] = "%{}%".format(post.get('author'))

    if post.get("repo_maintainer_id"):
        search += ["m.repo_maintainer_id = %(repo_maintainer_id)s"]
        args['repo_maintainer_id'] = post.get("repo_maintainer_id")

    if post.get("search"):
        # conditionary clause in this specific case
        name, shortdesc, summary, author, srch = map(unaccent, ['m.name', 'm.shortdesc', 'm.summary', 'm.author','%(srch)s'])

        search += [
            "({name} ilike {srch} OR {shortdesc} ilike {srch} OR {summary} ilike {srch} OR {author} ilike {srch})".format(
                name=name,
                shortdesc=shortdesc,
                summary=summary,
                author=author,
                srch= srch,
            )
        ]
        args['srch'] = "%{}%".format(post.get('search'))

    if maintainer_id:
        search += ["(m.is_published OR m.repo_maintainer_id=%(maintainer_id)s)"]
        args['maintainer_id'] = maintainer_id

    series = post.get('series', False)
    if series:
        join_clauses += ['JOIN loempia_series s ON s.id = m.series_id']
        join_clauses += ["JOIN loempia_module_template t ON t.id = m.template_id"]
        search += ["s.name = %(series)s and s.active = true"]
        args["series"] = str(series)
    else:
        join_clauses += ["JOIN loempia_module_template t ON t.latest_module_id = m.id"]

    if search:
        search = 'WHERE ' + ' AND '.join(search)
    else:
        search = ''

    if join_clauses:
        join_clauses = '\n'.join(join_clauses)
    else:
        join_clauses = ''

    limit = int(limit) if limit else "ALL"
    offset = int(post.get('offset', 0))

    query = """
        SELECT m.id
        FROM loempia_module m
        {join_clause}
        {search}
        ORDER BY {order}
        LIMIT {limit}
        OFFSET {offset}
    """.format(join_clause=join_clauses, search=search, order=order, limit=limit, offset=offset)

    env.cr.execute(query, args)
    return [id for [id] in env.cr.fetchall()]

def sanitize_url(url):
    # This function returns an url respecting the ssh protocol
    url = url.replace(' ', '')
    if not parse_url:
        return url
    pattern = r"^(ssh|http)"
    if not re.match(pattern, url):
        url = url.replace(':', '/')  # Otherwise git@github.com:path#version triggers an error
    parsed_url = parse_url(url)
    new_params = {'scheme': 'ssh'}
    scheme = parsed_url.scheme
    new_params['path'] = re.sub(r'(?<!\.git)$', '.git', parsed_url.path.rstrip('/'))
    auth = new_params['auth'] = parsed_url.auth or 'git'

    if scheme != 'ssh':
        new_params['auth'] = 'git'
    elif ':' in auth:
        new_params['auth'] = auth.split(':')[0]  # Remove the password if present in the auth

    sanitized_url = parsed_url._replace(**new_params).url
    if parsed_url.host == "ssh.dev.azure.com":
        sanitized_url = f"{new_params['auth']}@{parsed_url.host}:{parsed_url.path.lstrip('/')}#{parsed_url.fragment}"
    return sanitized_url

def get_values_from_manifest(path, readme_path, repo, supported_currencies):
    manifest_path = [os.path.join(path, x) for x in repo.series_id.manifest_file.split(',') if os.path.isfile(os.path.join(path, x))]
    if not manifest_path:
        raise ValueError('invalid manifest path')

    _logger.debug('create from path: %r on %s repo %s', path, repo.vcs, repo.url)
    with open(manifest_path[0]) as fp:
        try:
            manifest = literal_eval(fp.read())
        except Exception as e:
            raise ManifestError(u'Invalid manifest: %s' % (exception_to_unicode(e),))

    if not manifest.get('installable', True) and not (repo.is_server or repo.is_odoo_enterprise):
        raise ManifestError('Module not installable')

    name = os.path.basename(path)
    if name == 'base':
        raise ValueError('Reserved module name: base')
    version = manifest.get('version', '1.0')
    if version == repo.series_id.name or not version.startswith(repo.series_id.name + '.'):
        version = '%s.%s' % (repo.series_id.name, version)

    support = manifest.get('support', '')
    if not support:
        support = 'apps@odoo.com' if repo.auto_approve or not repo.maintainer_id or '@' not in repo.maintainer_id.login else repo.maintainer_id.login

    description = manifest.get('description', '')
    if not description:
        if readme_path:
            with open(readme_path[0]) as fp:
                description = fp.read()

    categories = manifest.get('category', 'Extra tools').split('/')
    lowercase_categories = [c.lower() for c in categories]

    price = manifest.get('price', 0.0)
    try:
        price = max(0.0, float(price))
    except ValueError:
        raise ManifestError('Invalid price set in manifest: %s' % (price))

    currency = manifest.get('currency', 'EUR').upper()
    if price > 0.0 and currency not in supported_currencies:
        raise ManifestError('Price set in manifest but not "currency". Accepted values are "EUR" and "USD".')

    author = manifest.get('author', ALIASES[0] if repo.auto_approve else 'Unknown')
    if isinstance(author, (str, bytes)):
        author = author.split(',')
    if isinstance(author, Iterable):
        author = [a.strip() if a.strip() not in ALIASES else ALIASES[0] for a in author]
    else:
        author = ['Unknown']
    author = ', '.join(author)

    return {
        'name': name,
        'version': version,
        'description': description,
        'shortdesc': manifest.get('name').strip() if manifest.get('name') else name.strip(),
        'author': author,
        'images': manifest.get('images', []),
        'maintainer': manifest.get('maintainer', False),
        'website': "http://{}".format(manifest.get('website'))
                   if manifest.get('website') and not url_parse(manifest.get('website', '')).scheme
                   else manifest.get('website', ''),
        'license': manifest.get('license') or 'Other proprietary',
        'application': manifest.get('application', False),
        'auto_install': manifest.get('auto_install', False),
        'summary': manifest.get('summary', ''),
        'live_test_url': manifest.get('url', '') or manifest.get('live_test_url', ''),
        'support': support,
        'categories': categories,
        'depends': manifest.get('depends', []),
        'is_theme': lowercase_categories[0] in ['theme', 'themes'],
        'is_hidden': 'hidden' in lowercase_categories,
        'price_single': price,
        'currency': currency,
    }


def _count_lines(module_path):
    module_path = module_path.rstrip('/')
    exclude_list = []
    for i in MANIFEST_NAMES:
        manifest_path = os.path.join(module_path, i)
        try:
            exclude_list.extend(cloc.DEFAULT_EXCLUDE)
            d = literal_eval(pathlib.Path(manifest_path).read_bytes().decode('latin1'))
            for j in ['cloc_exclude', 'demo', 'demo_xml']:
                exclude_list.extend(d.get(j, []))
            break
        except FileNotFoundError:
            pass
    exclude = set()
    for i in filter(None, exclude_list):
        exclude.update(str(p) for p in pathlib.Path(module_path).glob(i))

    cloc_obj = cloc.Cloc()
    qweb_views = {}
    qweb_exclude = []
    server_actions = {}
    server_exclude = []
    computed_fields = {}
    computed_exclude = []

    def get_id(element):
        id_str = element.get('id')
        if id_str is None:
            return ''.join(random.choice(string.ascii_lowercase) for _ in range(8))
        return id_str.split('.')[-1]

    def get_view_lines(element):
        inner = ''.join(etree.tostring(child, encoding='unicode') for child in element.iterchildren())
        return cloc_obj.parse_xml(inner + element.text)[0]

    def get_python_lines(element, field):
        code = element.find('field[@name="%s"]' % field)
        if code is not None:
            return cloc_obj.parse_py(code.text)[0]
        return 0

    def get_standard_cloc(file):
        # Decode using latin1 to avoid error that may raise by decoding with utf8
        # The chars not correctly decoded in latin1 have no impact on how many lines will be counted
        content = pathlib.Path(file).read_bytes().decode('latin1')
        return cloc_obj.parse(content, ext)[0]

    def parse_root(root_xml):
        for el in root_xml:
            if el.tag in ['odoo', 'data', 'openerp']:
                parse_root(el)
            elif el.tag == "template":
                qweb_views[get_id(el)] = get_view_lines(el)
            model = el.get('model')
            if model == 'ir.actions.server':
                state = el.find('field[@name="state"]')
                if state is not None and state.text == "code":
                    server_actions[get_id(el)] = get_python_lines(el, 'code')
            elif model == 'ir.model.fields':
                computed_fields[get_id(el)] = get_python_lines(el, 'compute')
            elif model == 'ir.ui.view':
                field_type = el.find('field[@name="type"]')
                if field_type is not None and field_type.text.strip() == 'qweb':
                    field_arch = el.find('field[@name="arch"]')
                    if field_arch is not None:
                        qweb_views[get_id(el)] = get_view_lines(field_arch)
            elif model == 'ir.model.data':
                module = el.find('field[@name="module"]')
                if module is not None and module.text == "__cloc_exclude__":
                    res_id = el.find('field[@name="res_id"]')
                    if res_id is not None:
                        res_id = res_id.get('eval') or ""
                        ids = re.split("'|\"", res_id)
                        ids = [_id for _id in ids if _id and '(' not in _id and ')' not in _id]
                        ids = [_id.split('.')[-1] for _id in ids]
                    model_field = el.find('field[@name="model"]')
                    if model_field is None:
                        continue
                    elif model_field.text == 'ir.actions.server':
                        server_exclude.extend(ids)
                    elif model_field.text == 'ir.model.fields':
                        computed_exclude.extend(ids)
                    elif model_field.text == 'ir.ui.view':
                        qweb_exclude.extend(ids)

    lines_of_code = 0
    for root, _dirs, files in os.walk(module_path):
        for file_to_count in files:
            ext = os.path.splitext(file_to_count)[1].lower()
            if ext not in cloc.VALID_EXTENSION:
                continue
            file_path = os.path.join(root, file_to_count)
            if file_path in exclude:
                continue
            if ext != '.xml':
                lines_of_code += get_standard_cloc(file_path)
                continue
            doc = etree.parse(file_path)
            root_elem = doc.getroot()
            parse_root(root_elem)
    for exclusion in server_exclude:
        server_actions.pop(exclusion, 0)
    for exclusion in computed_exclude:
        computed_fields.pop(exclusion, 0)
    for exclusion in qweb_exclude:
        qweb_views.pop(exclusion, 0)
    lines_of_code += sum(server_actions.values()) + sum(computed_fields.values()) + sum(qweb_views.values())
    return lines_of_code


def path_cleaner(values, paths):
    for path in paths:
        if values.get(path):
            values[path] = values[path].strip(r'\/')
    return values
